from setuptools import setup
import os

version = "0.1.0"
tag = os.getenv("TRAVIS_TAG")
assert not tag or tag == version, "Travis tag has to match version: %r!=%r" % (tag, version)

setup(
    name="legoboost",
    version=version,

    author="Daniel Ersen",
    author_email="danielersen@gmail.com",
    license="MIT",
    description="Python library Lego Boost forked by the pylgbst repository but the accelerometer is a new function.",
    url='https://github.com/danielersen/Lego-Boost',
    keywords=['LEGO', 'ROBOTICS', 'BLUETOOTH'],

    packages=["pylgbst", "pylgbst.comms"],
    requires=[],
    extras_require={
        # Note that dbus and gi are normally system packages
        "gatt": ["gatt", "dbus", "gi"],
        "gattlib": ["gattlib"],
        "pygatt": ["pygatt", "pexpect"],
        "bluepy": ["bluepy"],
        "bleak": ["bleak"],
    },
)
